/*
 * Created on May 3, 2005
 *
 * Last Modified on Nov. 27, 2005
 */
/**
 * @author zAlbee
 */
import java.io.*;
import java.util.*;
import java.text.*;

public class immerge {

	static final String NL = System.getProperty("line.separator"); // newline (CR/CRLF/LF)
	static final char[] UTF8BOM = {0xEF, 0xBB, 0xBF};
	static int tz_adjust = -16;

	/***
	 * General Log class, associated with a BufferedReader
	 *
	 * Subclasses need to implement:
	 *
	 *   readSess() - reads the next session into a buffer
	 *              - puts that session's date into the sessDate property.
	 *              - returns True if a session was read, False if the log was empty (no session read)
	 *
	 *   getBuffer() - should return all the text within that session.
	 */
	static abstract class Log implements Comparable {
		static final String sessStart  = "Session Start (";
		static final String sessStart2 = "sion Start (";
		static final String sessClose  = "Session Close (";
		static SimpleDateFormat sdfTrillian = new SimpleDateFormat("EEE MMM dd HH:mm:ss yyyy", Locale.US);
		static SimpleDateFormat sdfICQ = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss", Locale.US);
		static SimpleDateFormat sdfTTime = new SimpleDateFormat("[hh:mm a]", Locale.US);
		static SimpleDateFormat sdfTTime24 = new SimpleDateFormat("[HH:mm]", Locale.US);

		String filename;
		String pathname;
		BufferedReader reader;
		Date sessDate;
		String buffer;
		boolean empty = false;

		abstract boolean readSess() throws Exception;
		abstract String getBuffer();

		/** Log() contructor
		 */
		Log(String fname, String pname, BufferedReader br) throws IOException {
			/*
			sdfTrillian.setTimeZone(TimeZone.getTimeZone("UTC"));
			sdfICQ.setTimeZone(TimeZone.getTimeZone("UTC"));
			sdfTTime.setTimeZone(TimeZone.getTimeZone("UTC"));
			*/
			filename = fname;
			pathname = pname;
			reader = br;
			reader.mark(16);
			if (reader.read() == 0xEF && reader.read() == 0xBB && reader.read() == 0xBF) { 
				// UTF-8 header: ﻿
				//System.out.println("File " + filename + " starts with a UTF-8 BOM signature.");
			}
			else reader.reset();
		}

		public int compareTo(Object o2) {
			Log log2 = (Log)o2;
			return sessDate.compareTo(log2.sessDate);
		}

	}

	/***
	 * TrLog class, subclass of Log
	 * Represents a Trillian logfile.
	 */
	static class TrLog extends Log {
		boolean sessOpen = false;
		boolean startOfFile = true;

		/** TrLog() contructor
		 */
		TrLog(String fname, String pname, BufferedReader br) throws IOException {
			super(fname, pname, br);
		}

		/** TrLog.readSess()
		 */
		boolean readSess() throws Exception {
			sessOpen = false;
			String s = reader.readLine();

			int nameEndIdx;

			while (s != null && !sessOpen) {
				if (    ( s.startsWith( sessStart ) || (startOfFile && s.startsWith(sessStart2)) )
					 && ( nameEndIdx=s.indexOf("):", s.length()-28) ) > -1
					) {
					int dateStartIdx = nameEndIdx + 3;

					sessDate = sdfTrillian.parse(s.substring(dateStartIdx));

					if (s.startsWith("sion")) {
						s = "Ses" + s;
					}
					sessOpen = true;
					buffer = s + NL;
				}
				else if (s.trim().length()>0) {
					// Text exists outside a session. Write it anyway to keep fidelity of logs.
					// Give it a "dummy" session, with the same date as the last one, so that it's written right away.
					// Since there will be no session close marker, Read until the next session starts, then rewind the reader.
					
					/*
					Calculation of max rewind for   BufferedReader.mark(int readAheadLimit)
					
					"Session Start (email@hotmail.com:MY LONG MSN DISPLAY NAME): Sun Jul 25 00:54:39 2004"
					"Session Start (".length() == 15;
					"): Sun Jul 25 00:54:39 2004".length() == 27
					Email address max == 64 + '@' + 255
						  		  	  == 320
					MSN name max == 129;
					----
					Total == 15 + 320 + ":" + 129 + 27
						  == 492
					*/
					
					buffer = s + NL;
					reader.mark(512);
					s = reader.readLine();
					
					while (s != null) {
						if ( s.startsWith( sessStart ) && s.indexOf("):", s.length()-28) > -1 ) {
							reader.reset();
							return true;
						}
						buffer += s + NL;
						reader.mark(512);
						s = reader.readLine();
					}
					
					startOfFile = false;
					return true;
					
					//System.out.println("Warning: Skipped line = " + s);
				}
				
				startOfFile = false;
				s = reader.readLine();
			}

			while (s != null && sessOpen) {
				buffer += s + NL;

				if ( s.startsWith( sessClose ) && ( nameEndIdx=s.indexOf("):", s.length()-28) ) > -1 ) {
			  		sessOpen = false;
			  		buffer += NL + NL;
			  		break;
				}
				s = reader.readLine();
			}

			if (s==null) {
				if (sessOpen) ; // not terminated correctly. write it anyway.
				else {
					empty = true;
					reader.close();
					return false;
				}
			}
			return true;

		}

		String getBuffer() {
			return buffer;
		}
	} //end class TrLog


	/***
	 * HtmlLog class, subclass of Log
	 * Represents an ICQ log file (ICQ History Reader HTML format)
	 */
	static class HtmlLog extends Log {
		String yourNick;
		String nickName;
		int yourUIN;
		String theirUIN;
		String sendColour;
		String recvColour;
		static final String genStart = "<font size=3 color=";
		static final String genStart2 = ">";
		static final String genEnd = "</font>";
		static final int START_LEN = genStart.length() + 8;
		String sendStart;
		String recvStart;

		String buffer2;
		Date buffer2Date = null;

		/** HtmlLog() contructor
		 */
		HtmlLog(String fname, String pname, BufferedReader br) throws IOException {
			super(fname, pname, br);
		}
		
		void init() throws IOException{
			String s = reader.readLine();

			while (s != null) {
				if (s.matches("<td [^>]*>Nickname:</td>")) {
					s=reader.readLine();
					nickName = s.substring( s.indexOf('>') +1, s.lastIndexOf("</td>") );
					break;
				}
				else if (s.trim().toLowerCase().equals("no info for this user")) {
					System.out.println("Info: No nickname in " + filename);
					nickName = "" + theirUIN;
					s = reader.readLine();
					break;
				}
				s=reader.readLine();
			}

			if (s == null) {
				System.out.println("Error: Reached end-of-file while looking for info in " + filename);
				empty = true;
				return;
			}

			while (s != null) {
				if (s.matches("<tr><td [^>]*>Sent</td><td.*bgcolor=.*")) {
					int colIdx = s.indexOf("bgcolor=") + 8;
					sendColour = s.substring( colIdx, colIdx + 7);
					sendStart = genStart + sendColour + genStart2;
					break;
				}
				s=reader.readLine();
			}

			if (s == null) {
				System.out.println("Error: Could not read sent colour from " + filename);
				empty = true;
				return;
			}

			while (s != null) {
				if (s.matches("<tr><td [^>]*>Recieved</td><td.*bgcolor=.*")) {
					int colIdx = s.indexOf("bgcolor=") + 8;
					recvColour = s.substring( colIdx, colIdx + 7);
					recvStart = genStart + recvColour + genStart2;
					break;
				}
				s=reader.readLine();
			}

			if (s == null) {
				System.out.println("Error: Could not read receive colour from " + filename);
				empty = true;
				return;
			}

		}

		/** HtmlLog.readMsg()
		 */
		String readMsg(Date[] dpoint) throws Exception {
			boolean msgOpen = false;
			boolean isSend;
			String msg = "";

			String s=reader.readLine();
			dpoint[0] = null;

			// grab date

			while (s != null && dpoint[0]==null) {
				if ( s.startsWith(genStart) ) {
					int dateStartIdx = s.indexOf('(', START_LEN) + 1;
					int dateEndIdx = s.lastIndexOf(')');
					dpoint[0] = sdfICQ.parse(s.substring(dateStartIdx, dateEndIdx));
					//dpoint[0].setTime( dpoint[0].getTime() - 16*60*60*1000 ); // has sync problems when dates are near daylight savings switch
					dpoint[0].setHours( dpoint[0].getHours() + tz_adjust); // -16 hours seems to make the time correct? no idea how ICQHR calculates timezone
				}
				s = reader.readLine();
			}

			while (s != null && !msgOpen) {
				if ( s.startsWith(genStart)) {
					msgOpen = true;
					isSend = s.startsWith(sendStart);


					msg = sdfTTime.format(dpoint[0]) + " ";
					if (isSend) msg += yourNick + ": ";
					else msg += nickName + ": ";

					if (s.trim().endsWith(genEnd)) {
						msg += s.substring( START_LEN, s.lastIndexOf(genEnd)) + NL; msgOpen = false; break;}
					else if (s.trim().endsWith("<BR>")) msg += s.substring( START_LEN, s.lastIndexOf("<BR>") ) + NL;
					else msg += s.substring( START_LEN ) + NL;
				}

				s=reader.readLine();
			}

			while (s != null && msgOpen) {
				if (s.trim().endsWith(genEnd)) {
					msg += s.substring( 0, s.lastIndexOf(genEnd)) + NL; msgOpen = false; break;}
				else if (s.trim().endsWith("<BR>")) msg += s.substring( 0, s.lastIndexOf("<BR>") ) + NL;
				else msg += s + NL;

				s=reader.readLine();
			}

			return msg;
		}

		/** HtmlLog.readSess()
		 */
		boolean readSess() throws Exception {
			Date sDate = buffer2Date;
			Date eDate;
			String m = null;
			Date[] dp = new Date[] {null};

			if (sDate == null) {
				m = readMsg(dp);
				if (m.length()==0) {empty=true; reader.close(); return false;}
				sDate = dp[0];
			}
			else {
				m = buffer2;
				buffer2 = null;
				buffer2Date = null;
			}

			sessDate = eDate = sDate;
			buffer = sessStart + (yourUIN>0 ? ""+yourUIN:"ICQ") + ":" + nickName + "): " + sdfTrillian.format(sDate) + NL + unHTMLEntity(m);

			m = readMsg(dp);

			while (m.length() > 0) {
				Date date = dp[0];
				if (date.getTime() < sDate.getTime() + 24*60*60*1000) {
					buffer += unHTMLEntity(m);
					eDate = date;
				}
				else {
					buffer2 = m;
					buffer2Date = date;
					break;
				}

				m = readMsg(dp);
			}

			buffer += sessClose + nickName + "): " + sdfTrillian.format(eDate) + NL+NL+NL;

			return true;
		}

		String getBuffer() {
			return buffer;
		}

	} //end class HtmlLog


	/**
	 * unHTMLEntity ( String )
	 *
	 * Returns a new string with various HTML escape sequences (e.g. &amp;) removed
	 * and replaced with the normal characters.
	 */
	static String unHTMLEntity(String s) {
		return s.replaceAll("&amp;","&").replaceAll("&nbsp;"," ").replaceAll("&lt;","<").replaceAll("&gt;",">");
	}

	/**
	 * readIntFrom ( BufferedReader )
	 *
	 * Reads an integer from the stream, returns it.
	 * Will reprompt and retry the read until an integer is found.
	 * Throws IOException if an I/O error occurs.
	 */
	static int readIntFrom(BufferedReader reader) throws IOException{
		String utxt = reader.readLine();
		try {
			int i = Integer.parseInt(utxt.trim());
			return i;
		} catch (Exception e) {
			System.out.print("You must enter a number: ");
			return readIntFrom(reader);
		}
	}
	
	static boolean readYesNoFrom(BufferedReader reader) throws Exception{
		boolean yesno = false;
		String utxt = reader.readLine().trim() + " ";
		char first = utxt.charAt(0);
		if (first == 'Y' || first == 'y') return true;
		else if (first == 'N' || first == 'n') return false;
		else {
			System.out.print("Yes or No? ");
			return readYesNoFrom(reader);
		}
	}

	/***
	 *  main ( String[] )
	 *
	 *  Asks the user for some information.
	 *  Then executes the merging of Trillian logs.
	 ***/
	public static void main(String[] args) throws Exception {
		int numTrLog, numHtml, yourUIN = 0;
		String yourNick = null;

		String[] mediums = {"AIM", "ICQ", "IRC", "MSN", "YAHOO"};
		String outPath = "output";
		String[] outPath2 = new String[5];

		File outDir = new File(outPath);
		if (!outDir.isDirectory() && !outDir.mkdir()) {
			System.out.println("Couldn't create '" + outPath + "' directory");
			System.out.println("Failed.");
			return;
		}

		BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
		System.out.print("How many folders (logs1 .. logsN) of Trillian-style <username>.log: ");
		numTrLog = readIntFrom(in);
		System.out.print("How many folders (icqh1 .. icqhN) of IcqHR output <UIN>.html (UIN per file): ");
		numHtml = readIntFrom(in);

		if (numHtml > 0) {
			System.out.print("Your ICQ UIN (if you don't know/care, enter 0): ");
			yourUIN = readIntFrom(in);
			System.out.print("Your ICQ Nickname (as you want it to appear): ");
			yourNick = in.readLine();
			System.out.print("Correct ICQHR timezone by how many hours? (-16 = default, 0 = no change):");
			tz_adjust = readIntFrom(in);
			System.out.print("Display time in 24-hr format? (yes = 24-hr, no = AM/PM): ");
			boolean use24h = readYesNoFrom(in);
			if (use24h) {
				Log.sdfTTime = Log.sdfTTime24;
			}
		}

		System.out.print("Should the logs be output to a \"Query\" folder? (Yes/No): ");
		boolean useQuery = readYesNoFrom(in);
		
		System.out.println();

		for (int j=0; j<mediums.length; j++) {
			System.out.println("== " + mediums[j] + " ==");

			// link the username to a Vector of the logs
			Hashtable userLogs = new Hashtable();

			// Find files - Trillian .log

			for (int i=0; i<numTrLog; i++) {
				String logsPath = "logs" + (i+1) + File.separator + mediums[j]; // in folder
				String queryPath = logsPath + File.separator + "Query";         // in Query folder
				
				File dir = new File(logsPath);
				File dirQ = new File(queryPath);
				if (!dir.isDirectory()) continue;

				File[][] files = new File[2][];
				files[0] = dir.listFiles();

				if (dirQ.isDirectory())
					files[1] = dirQ.listFiles();
				else
					files[1] = new File[0];
					
				for (int qd = 0; qd <= 1; qd++) {
					for (int k=0; k<files[qd].length; k++) {
						File f = files[qd][k];
						String filename = f.getName();
	
						if (f.isFile() && filename.endsWith(".log")) {
							TrLog lg = new TrLog( filename, f.getPath(), new BufferedReader(new FileReader(f)) );
							String user = filename.substring(0, filename.length()-4);
							if (userLogs.containsKey(user))
								((Vector)userLogs.get(user)).add(lg);
							else {
								Vector logs = new Vector();
								logs.add(lg);
								userLogs.put(user, logs);
							}
						}
					}
				}
				
			}

			// Find files - ICQ History Reader 1.8 .html output
			if (j==1) { // ICQ

				for (int i=0; i<numHtml; i++) {
					File dir = new File("icqh" + (i+1));
					if (!dir.isDirectory()) continue;

					File[] files;
					files = dir.listFiles();
					for (int k=0; k<files.length; k++) {
						String filename = files[k].getName();

						if (files[k].isFile() && filename.endsWith(".html")) {
							HtmlLog lg = new HtmlLog( filename, files[k].getPath(), new BufferedReader(new FileReader(files[k])) );
							String user = filename.substring(0, filename.length()-5);
							lg.yourNick = yourNick;
							lg.yourUIN = yourUIN;
							lg.theirUIN = user;
							lg.init();
							if (userLogs.containsKey(user))
								((Vector)userLogs.get(user)).add(lg);
							else {
								Vector logs = new Vector();
								logs.add(lg);
								userLogs.put(user, logs);
							}
						}
					}
				}
			}

			// if logs exist, make output dir
			if (userLogs.size() > 0) {
				outPath2[j] = outPath + File.separator + mediums[j] + (useQuery? File.separator+"Query" : "");
				
				File outDir2 = new File(outPath2[j]);
				if (!outDir2.isDirectory() && !outDir2.mkdirs()) {
					System.out.println("Couldn't create directory '" + outPath2[j] + "'");
					System.out.println("Skipping " + mediums[j]);
					continue;
				}
			}

			// Iterate through usernames.
			int users = 0;
			Enumeration userNames = userLogs.keys();

			while (userNames.hasMoreElements()) {
	  			users++;
				int sessions = 0;
				int dupes = 0;
				int numFiles;

				String user = (String)userNames.nextElement();
				Vector logs = (Vector)userLogs.get(user);

				// Initialize; grab one session from each log, and record in log buffer
				for (int i=0; i<logs.size(); i++) {
					Log lg = (Log)logs.get(i);
					lg.sessDate = new Date(i); // initialize date, in case the readSess() func doesn't find a session date.
					if (!lg.readSess()) logs.remove(i--);
				}
				Collections.sort(logs);
				numFiles = logs.size();

				File f = new File(outPath2[j] + File.separator + user + ".log");
				if (!f.exists()) f.createNewFile();

				try {
					FileWriter fw = new FileWriter(f);
					fw.write(UTF8BOM);

					while (logs.size()>0) {
						sessions++;
						Log lg = (Log)logs.get(0);
						fw.write(lg.getBuffer());

						for (int i=1; i<logs.size(); i++) {
							Log olog = (Log)logs.get(i);
							if (!lg.sessDate.equals(olog.sessDate)) {
								//System.out.println("No Dup - " + lg.sessDate + " vs " + olog.sessDate);
								break;
							}

							dupes++;
							//System.out.println("-- Duplicate in " + lg.filename + "(" + lg.pathname + " & " + olog.pathname + ") on " + lg.sessDate);
							if (!olog.readSess()) logs.remove(i--);
						}

						if (!lg.readSess()) logs.remove(0);
						else Collections.sort(logs);
					}

					fw.close();

					System.out.println(user + ": merged " + numFiles + " files, " + sessions + " sessions, " + dupes + " duplicates");
				}
				catch (IOException e) {
					e.printStackTrace();
				}

			} //end while enumeration of users

			System.out.println(users + " Files written.\n");

		} //end for mediums

		in.close();

		System.out.println("Finished.");

	} // end main

}
